(* $Id: cmd.ml 1167 2012-12-28 22:13:41Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_query
open Printf

exception Exit_code of int



let usage() =
  prerr_endline ("usage: godi_console [<command> [<options>]]\n" ^
		 "Without command: start interactive console\n" ^
		 "Available commands:\n" ^
		 "  update - Update the build instructions\n" ^
		 "  wish - Manipulate the list of user's wishes\n" ^
		 "  perform - Install or remove packages\n" ^
		 "  list - List packages by criterion\n" ^
		 "  info - Show details about installed package or tgz file\n" ^
                 "  dep - Check dependency on installed package\n" ^ 
		 "  delete - Remove packages (godi_delete backend)\n" ^
		 "  add - Install binary packages (godi_add backend)\n" ^ 
		 "  ptool - Process packages with the package tool\n" ^ 
		 "  profile - Manage profiles\n" ^
		 "  rprofile - Manage remote profiles\n" ^
                 "  initool - Manipulate package.ini files\n" ^
		 "  make - Invoke the built-in 'make' utility (godi_make backend)\n" ^ 
		 "  pax - Invoke the built-in 'pax' utility (godi_pax backend)\n" ^ 
		 "  wtool - Interact with remote locations (HTTP, FTP)\n" ^ 
		 "The commands have their own help texts, use -help\n" ^
		 "to see them, e.g.\n" ^
		 "  godi_console perform -help");
  exit 1
;;


let catch_error f arg =
  let flush() = flush stdout; flush stderr in
  try
    f arg
  with
    | Godi_plan.Mission_impossible text ->
	flush();
	prerr_endline "There is no way to perform the task.";
	prerr_endline text;
	raise (Exit_code 1)
    | Godi_plan.Package_installation_not_unique name ->
	flush();
	prerr_endline ("Tried to install several versions of this package at once: " ^ name);
	raise (Exit_code 1)
    | User_error msg ->
	flush();
	prerr_endline ("Error: " ^ msg);
	raise (Exit_code 1)
    | Godi_plist.Plist_error (file, line) ->
	flush();
	prerr_endline ("Error in PLIST file " ^ file ^ ", in line:\n" ^ line);
	raise (Exit_code 1)
    | Godi_plist.Plist_processing_error (file, text) ->
	flush();
	prerr_endline ("Error in PLIST file " ^ file ^ ": " ^ text);
	raise (Exit_code 1)
    | Godi_plist.Plist_expansion_error text ->
	flush();
	prerr_endline ("Error during PLIST expansion: " ^ text);
	raise (Exit_code 1)
    | Godi_make.Out_of_date ->
	(* Silent! *)
	raise (Exit_code 1)
    | Arg.Bad msg ->
	flush();
	prerr_endline msg;
	raise (Exit_code 1)
    | Arg.Help msg ->
	flush();
	prerr_endline msg;
	raise (Exit_code 0)
;;


let current_user() =
  try Sys.getenv "USER"
  with
    | Not_found ->
        try Sys.getenv "USERNAME"
        with
          | Not_found ->
              "nobody"


let opt_re = Str.regexp "^\\([^=:]+\\):\\([^=]+\\)=\\(.*\\)$";;
let ropt_re = Str.regexp "^\\([^=:]+\\):\\([^=]+\\)$";;

let split_re = Str.regexp "[ \t\r\n]+";;
let split s = Str.split split_re s;;

let word_by_word f s =
  let l = split s in
  List.iter f l
;;


let assign_re = Str.regexp "^\\([^=]+\\)=\\(.*\\)$";;

let decompose_assignment s =
  if Str.string_match assign_re s 0 then
    Some(Str.matched_group 1 s, Str.matched_group 2 s)
  else
    None
;;


let inst_pkg_by_name con name =
  let (n, version, rev) =
    Godi_versions.split_name_version name in
  let pkg =
    try
      get_this_installed_package con n
    with
      | Not_found ->
	  user_error ("No such installed package: " ^ n) in
  if version <> "" then (
    let dep = Godi_versions.parse_dependency (n ^ "==" ^ version) in
    if not (Godi_query.matches_dep pkg dep) then
      user_error ("Package not installed in requested version: " ^ name);
    (* We do not check the revision number... *)
  );
  pkg
;;


let helptext s =
  assert(!Arg.current <= 1);
  "usage: " ^ 
    (if !Arg.current = 0 then
       (Filename.basename Sys.argv.(0))
     else
       (Filename.basename Sys.argv.(0)) ^ " " ^ Sys.argv.(1)
    ) ^ " " ^ s
;;


let cmd_update() =
  let quick = ref false in
  Arg.parse 
      [ "-quick", Arg.Set quick,
	"  only extract buildfiles actually downloaded";
      ]
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      (helptext "[<options>]");
  let con = stderr_console() in
  ( try
      catch_error Godi_profile_ops.rupdate_current con;
      Godi_profile_ops.enable (Godi_query.profile_name())
    with
      | Exit_code _ ->
          prerr_endline "  [Ignoring error]"
  );
  let profile_ini =
    Godi_file.addname (Godi_query.profile_filename()) "profile.ini" in
  let profile =
    Godi_profile.read_profile_ini profile_ini in
  let names = Godi_update.update_buildfiles con profile in
  let extract_names = if !quick then Some names else None in
  Godi_update.extract_buildfiles ?names:extract_names con profile
;;


let cmd_wish() =
  let wishes = ref (Godi_wishes.read_wishes ~pref_user_error:true ()) in
  let modified = ref false in
  let con = stderr_console() in

  let check_src_pkg pkgname =
    let src_pkgs = scan_source_packages con in
    if not (List.exists (fun p -> pkg_name p = pkgname) src_pkgs) then
      user_error ("No such source package: " ^ pkgname)
  in

  let add_build_wish pkgname =
    check_src_pkg pkgname;
    let wish = `Build pkgname in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
      modified := true
    )
  in
 
  let build_all_installed() =
    let src_pkgs = scan_source_packages con in
    let src_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) src_pkgs) in
    let inst_pkgs = get_installed_packages con in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let spkgs = PkgMap.find_by ~name:n src_pkgs_map in
	 if spkgs <> [] then add_build_wish n
      )
      inst_pkgs
  in

  let rem_build_wish pkgname =
    let wish = `Build pkgname in
    wishes := List.filter (fun w -> w <> wish) !wishes;
    modified := true
  in
    
  let add_remove_wish pkgname =
    let pkg = inst_pkg_by_name con pkgname in
    let wish = `Remove (pkg_name pkg) in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
      modified := true
    );
  in
 
  let rem_remove_wish pkgname =
    let (name, _, _) = Godi_versions.split_name_version pkgname in
    let wish = `Remove name in
    wishes := List.filter (fun w -> w <> wish) !wishes;
    modified := true
  in
    
  let add_option opt =
    if Str.string_match opt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let optvalue = Str.matched_group 3 opt in
      check_src_pkg pkgname;
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl @ [ `Option(pkgname,optname,optvalue) ];
      modified := true
    )
    else
      user_error ("Syntax error in -option " ^ opt)
  in
    
  let rem_option opt =
    if Str.string_match ropt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl;
      modified := true
    )
    else
      user_error ("Syntax error in -no-option " ^ opt)
  in

  let reset() =
    wishes := [];
    modified := true
  in

  let print() =
    (* Sort wishes: *)
    let pkg_of_wish =
      function
	  `Build p -> p
        | `Add(p,_,_) -> p
	| `Remove p -> p
	| `Option(p,_,_) -> p in
    let cat_of_wish =
      function
	  `Build _ -> 1
        | `Add _ -> 2
	| `Remove _ -> 0
	| `Option(_,_,_) -> 3 in
    let optname_of_wish =
      function
	  `Build _ -> ""
	| `Add _ -> ""
	| `Remove _ -> ""
	| `Option(n,_,_) -> n in
    let cmp w1 w2 =
      let p1 = pkg_of_wish w1 in
      let p2 = pkg_of_wish w2 in
      match compare p1 p2 with
	  0 ->
	    ( match cat_of_wish w1 - cat_of_wish w2 with
		  0 ->
		    compare (optname_of_wish w1) (optname_of_wish w2)
		| n -> n
	    )
	| n -> n
    in
    let wl = List.sort cmp !wishes in
    (* Finally print wishes: *)
    List.iter
      (function
	 | `Build p ->
	     printf "-build  %s\n" p
         | `Add(p,_,_) ->
             printf "-add %s\n" p   (* TODO: implement -add *)
	 | `Remove p ->
	     printf "-remove %s\n" p
	 | `Option(p,n,v) ->
	     printf "-option %s:%s=%s\n" p n v
      )
      wl
  in

  Arg.parse 
      [ "-build", Arg.String (word_by_word add_build_wish),
	       "<pkg>                 Add wish to build <pkg>";
	"-no-build", Arg.String (word_by_word rem_build_wish),
	          "<pkg>              Withdraw wish to build <pkg>";
	"-remove", Arg.String (word_by_word add_remove_wish),
	        "<pkg>                Add wish to remove <pkg>";
	"-no-remove", Arg.String (word_by_word rem_remove_wish),
	           "<pkg>             Withdraw wish to remove <pkg>";
	"-option", Arg.String add_option,
	        "<pkg>:<name>=<val>   Add option <name>=<val> for <pkg>";
	"-no-option", Arg.String rem_option,
	           "<pkg>:<name>      Withdraw option <name> for <pkg>";
	"-rebuild", Arg.Unit build_all_installed,
	         "                    Wish to rebuild all installed packages";
	"-reset", Arg.Unit reset,
	       "                      Withdraw all wishes";
	"-print", Arg.Unit print,
	       "                      Print resulting wish list";
      ]
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      (helptext "<options>");

  if !modified then
    Godi_wishes.write_wishes ~pref_user_error:true !wishes
;;


let cmd_profile() =
  let con = stderr_console() in
  let mode = ref `Current in
  let opt_update = ref false in
  let opt_wish = ref false in
  let constraints = ref [] in
  Arg.parse
    [ "-list", Arg.Unit (fun () -> mode := `List),
            "           list available profiles";

      "-current", Arg.Unit (fun () -> mode := `Current),
               "        show the name of the current profile";

      "-enable", Arg.String (fun s -> mode := `Enable s),
              "<name>   enable this profile";

      "-update", Arg.Set opt_update,
              "         with -enable: also update pkgs from remote server";

      "-wish", Arg.Set opt_wish,
            "           with -enable: also set wishes for rebuild";

      "-where", Arg.String (fun s ->
                              let dep = 
                                Godi_versions.parse_dependency ~sloppy:true s in
                              constraints := dep :: !constraints
                           ),
      "<pkgcond>  Add ad-hoc constraint for package version selection ";
      "-delete", Arg.String (fun s -> mode := `Delete s),
              "<name>   delete this profile";
      
      "-copy-to", Arg.String (fun s -> mode := `Copy s),
               "<name>  copy the current profile, and name the copy";

      "-save-patches", Arg.Unit (fun () -> mode := `Save_patches),
                    "   save the profile patches in profile.ini";
    ]
    (fun _ -> raise(Arg.Bad "Unexpected argument"))
    (helptext "<options>");

  match !mode with
    | `List ->
	let items = Godi_profile_ops.commented_list() in
	List.iter
          (fun (n,c) ->
             printf "%-15s %s\n" n c
          )
          items
    | `Current ->
	print_endline (Godi_query.profile_name())
    | `Enable n ->
	Godi_profile_ops.enable n;
	if !opt_update then Godi_profile_ops.update con;
	Godi_profile_ops.extract ~constraints:!constraints con;
	if !opt_wish then Godi_profile_ops.wish con
    | `Delete n ->
	Godi_profile_ops.delete n
    | `Copy dest ->
	let src = Godi_query.profile_name() in
	Godi_profile_ops.copy src dest
    | `Save_patches ->
        Godi_profile_ops.save_patches con


let configure_wtool wtool timeout user =
  wtool # configure_timeout timeout;
  wtool # configure_store_passwords (Godi_query.store_passwords());
  wtool # configure_localbase (Godi_query.localbase());
  ( match user with
      | None ->
	  ( match Godi_query.user() with
	      | None -> ()
	      | Some u -> wtool # configure_user u
	  )
      | Some u ->
	  wtool # configure_user u
  )


let cmd_rprofile() =
  (* let con = stderr_console() in *)
  let mode = ref `None in
  let site = ref None in
  let timeout = ref 30.0 in
  let user = ref None in
  Arg.parse
    [ "-get", Arg.String (fun s -> mode := `Get s),
           "<name>    Retrieve this profile from a remote server";

      "-update", Arg.String (fun s -> mode := `Update s),
           "<name>    Retrieve if a newer version is available";

      "-put", Arg.String (fun s -> mode := `Put s),
           "<name>    Upload this profile to a remote server";

      "-list", Arg.Unit (fun () -> mode := `List),
            "         List available remote profiles";

      "-site", Arg.String (fun s -> site := Some s),
            "<url>    Set the remote directory for -get, -put, -list";

      "-user", Arg.String (fun s -> user := Some s),
      "<name>         Substitute this user for 'GODI_USER' in authentication";

      "-timeout", Arg.Set_float timeout,
      "<seconds>      Set the timeout to this value";
    ]
    (fun _ -> raise(Arg.Bad "Unexpected argument"))
    (helptext "<options>");

  let wtool = Wtool.wtool_factory() in
  configure_wtool wtool !timeout !user;
  let get_site () = 
    match !site with
      | None -> failwith "No -site set"
      | Some url -> url in
  match !mode with
    | `None ->
        failwith "No mode selected"
    | `Get n ->
        Godi_profile_ops.rget wtool (get_site()) n
    | `Update n ->
        Godi_profile_ops.rupdate wtool n
    | `Put n ->
        Godi_profile_ops.rput wtool (get_site()) n
    | `List ->
        let names = Godi_profile_ops.rlist wtool (get_site()) in
	List.iter print_endline names


let local_or_remote_file con fn =
  try
    let dir, base = Godi_file_handler.split_name fn in
    let hdl = Godi_file_handler.handler_of_directory dir in
    let real_fn, cleanup = hdl#get con base in
    (real_fn, cleanup)
  with
    | File_handler_error msg ->
	raise(User_error("Cannot get " ^ fn ^ ": " ^ msg))
;;


let cmd_perform() =
  let con = stderr_console() in
  let wishes = ref [] in
  let use_wishlist = ref false in
  let dry_run = ref false in
  let build_all_installed = ref false in
  let newer = ref false in
  let binary_additions = ref [] in
  let disable_builder = ref false in
  let replace = ref `Upgrade in
  let as_much_as_possible = ref false in

  let disabled_pkg = Hashtbl.create 10 in
    (* In as_much_as_possible mode, these are the packages that are disabled
       because of errors
     *)

  let check_src_pkg pkgname =
    let src_pkgs = scan_source_packages con in
    if not (List.exists (fun p -> pkg_name p = pkgname) src_pkgs) then
      user_error ("No such source package: " ^ pkgname)
  in

  let add_build_wish pkgname =
    check_src_pkg pkgname;
    let wish = `Build pkgname in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
    )
  in
 
  let add_remove_wish pkgname =
    let pkg = inst_pkg_by_name con pkgname in
    let wish = `Remove (pkg_name pkg) in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
    );
  in
 
  let add_option opt =
    if Str.string_match opt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let optvalue = Str.matched_group 3 opt in
      check_src_pkg pkgname;
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl @ [ `Option(pkgname,optname,optvalue) ];
    )
    else
      user_error ("Syntax error in -option " ^ opt)
  in

  let add_wishes() =
    use_wishlist := true
  in

  let add_package file =
    binary_additions := !binary_additions @ [fst(local_or_remote_file con file)]
  in

  Arg.parse 
      [ "-build", Arg.String (word_by_word add_build_wish),
	       "<pkg>                 Build <pkg>";
	"-remove", Arg.String (word_by_word add_remove_wish),
	        "<pkg>                Remove <pkg>";
	"-add", Arg.String add_package,
	     "<tgzfile>               Add binary package <tgzfile> (may be a URL)";
	"-option", Arg.String add_option,
	        "<pkg>:<name>=<val>   Set option <name>=<val> for <pkg>";
	"-wishes", Arg.Unit add_wishes,
	        "                     Build/remove packages of the wish list";
	"-rebuild", Arg.Set build_all_installed,
	         "                    Rebuild all installed packages";
	"-newer", Arg.Set newer,
	       "                      Install only newer packages";
	"-disable-builder", Arg.Set disable_builder,
	                 "            Do not try to build packages";
        "-no-upgrades", Arg.Unit (fun () -> replace := `Never),
                     "                Never upgrade packages to fulfill deps";
(*
	"-auto-replace-old", Arg.Unit (fun _ -> replace := `Upgrade),
                          "           Try to upgrade old packages if necessary";
 *)
	"-dry-run", Arg.Set dry_run,
	         "                    Only generate a plan but don't carry it out";
	"-tmpdir-state", Arg.String Godi_file.init_tmpdir_state,
	              "               (internal option)";
	"-as-much-as-possible", Arg.Set as_much_as_possible,
	                     "        Try to continue after build errors";
      ]
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      (helptext "<options>");

  (* First scan packages *)
  let src_pkgs = 
    if !disable_builder then [] else scan_source_packages con in
  let inst_pkgs = get_installed_packages con in
  let bin_pkgs = List.map get_binary_package !binary_additions in

  (* Apply wishes *)
  Godi_wishes.apply_wishes_to_packages src_pkgs !wishes;
  Godi_wishes.apply_wishes_to_packages inst_pkgs !wishes;
  List.iter (fun pkg -> (any_pkg pkg) # set_user_status `Add) bin_pkgs;

  let persistent_wl = Godi_wishes.read_wishes ~pref_user_error:true () in
  if !use_wishlist then (
    Godi_wishes.apply_wishes_to_packages src_pkgs persistent_wl;
    Godi_wishes.apply_wishes_to_packages inst_pkgs persistent_wl;
    (* wishlist does not support binary packages *)
  );
 
  if !build_all_installed then (
    let src_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) src_pkgs) in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let spkgs = PkgMap.find_by ~name:n src_pkgs_map in
	 List.iter
	   (fun (spkg,_) ->
	      (source_pkg spkg)#set_user_status `Build
	   )
	   spkgs
      )
      inst_pkgs
  );

  if !newer then (
    let inst_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) inst_pkgs) in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let ipkgs = PkgMap.find_by ~name:n inst_pkgs_map in
	 match ipkgs with
	     [] ->
	       ()
	   | [ipkg,_] ->
	       (* Is pkg newer than ipkg? *)
	       if not(newer_package pkg ipkg) then (
		 (* pkg is not newer, so do not build it *)
		 (any_pkg pkg)#set_user_status `Keep
	       )
	   | _ ->
	       (* Several installed packages with the same name: impossible *)
	       assert false
      )
      (src_pkgs @ bin_pkgs)
  );

  (* Prepare a function to write wishes back to the file: *)
  let save_wishes() =
    if !use_wishlist then (
      let wl = Godi_wishes.wishes_of_packages(src_pkgs @ inst_pkgs) in
      (* We only write those wishes that are also contained in 
       * persistent_wl. - Binary packages are out of scope.
       *)
      let wl' =
	List.filter (fun w -> List.mem w persistent_wl) wl in
      Godi_wishes.write_wishes ~pref_user_error:true wl';
    )
  in

  (* Write a temporary godi.conf file: *)
  let tmp = Godi_file.allocate_tmpdir() in
  let tmp_path = Godi_file.tmpdir_path tmp in
  let tmp_godi_conf = Godi_file.addname tmp_path "godi.conf" in
  let mflag = Godi_conffile.modify_godi_conf
                ~pref_user_error:true ~destfile:tmp_godi_conf src_pkgs in
 
  (* When options have been modified, rescan source packages: *)
  let vid = new_var_id() in
  let src_pkgs' =
    if mflag then (
      List.iter (fun pkg -> 
		   (source_pkg pkg)#rescan_deps con 
		     (Godi_file.slashname tmp_godi_conf) vid)
		src_pkgs;
      List.map 
	(fun pkg -> `Source_pkg ((source_pkg pkg)#variant vid))
	src_pkgs
    )
    else src_pkgs in
  Godi_file.release_tmpdir tmp;

  (* Generate the plan: *)
  let _n = Godi_plan.update_plan 
	    ~replace:!replace 
	    ~iteration:1 
	    (src_pkgs' @ inst_pkgs @ bin_pkgs) in
  (* May raise Mission_impossible, or Package_installation_not_unique *)
 
  let to_remove =
    List.filter to_be_removed inst_pkgs in

  let to_remove' =
    List.rev(Godi_plan.sort_packages to_remove) in
 
  let to_build_or_add =
    List.filter to_be_built src_pkgs' @
    List.filter to_be_added bin_pkgs in

  let to_build_or_add' =
    Godi_plan.sort_packages to_build_or_add in

  (* Execute the plan: *)

  if to_remove = [] && to_build_or_add = [] && not mflag then
    print_endline "Nothing to do."
  else (
    List.iter
      (fun pkg ->
	 if pkgtype pkg = `Source then (
	   printf "### Fetching %s (if necessary)\n" (pkg_name pkg);
	   flush stdout;
	   if not !dry_run then
	     (* TODO: in as_much_as_possible mode catch errors *)
	     Godi_perform.fetch_pkg ~pref_user_error:true con pkg;
	 )
      )
      to_build_or_add';
    List.iter
      (fun pkg ->
	 if (installed_pkg pkg)#preserve then
	   printf "### Package %s cannot be removed, skipping\n" (pkg_fullname pkg)
	 else (
	   printf "### Removing %s\n" (pkg_fullname pkg);
	   flush stdout;
	   if not !dry_run then (
	     ( try
		 Godi_perform.remove_pkg ~soft_errors:true con pkg
	       with
		   Godi_perform.Soft_rm_error ->
		     ()
	     );
	     (installed_pkg pkg)#set_user_status `Keep;
	     save_wishes();
	   )
	 )
      )
      to_remove';
    
    if mflag then (
      printf "### Modifying godi.conf\n";
      flush stdout;
      if not !dry_run then (
	let _ = 
	  Godi_conffile.modify_godi_conf ~pref_user_error:true to_build_or_add' 
	in ()
      )
    );

    List.iter
      (fun pkg ->
	 match pkgtype pkg with
	   | `Source 
	   | `Wrapped ->
	       let n = pkg_name pkg in
	       if Hashtbl.mem disabled_pkg n then
		 printf "### Skipping build of %s because of build errors in previous packages\n" n
	       else (
		 printf "### Building %s\n" n;
		 flush stdout;
		 if not !dry_run then (
		   try
		     Godi_perform.build_pkg ~pref_user_error:true con pkg;
		     (any_pkg pkg)#set_user_status `Keep;
		     save_wishes();
		   with
		     | Restart_recommended sh_cmd ->
			 printf "### The installation of the package requires a restart of this program!\n";
			 flush stdout;
			 (* Remove the wish, if any. Note that we do not know whether
			  * sh_cmd will be successful, so this may be too early.
			  *)
			 (any_pkg pkg)#set_user_status `Keep;
			 save_wishes();
			 (* Determine the remaining packages to build: *)
			 let to_do = ref [] in
			 let found = ref false in
			 List.iter (fun pkg' -> 
				      if !found then to_do := pkg' :: !to_do;
				      if pkg = pkg' then found := true) 
			   to_build_or_add';
			 let restart_args =
			   List.map
			     (fun pkg' ->
				match pkg' with
				    `Source_pkg _ ->
				      "-build " ^ Godi_file.uquote (pkg_name pkg')
				  | `Binary_pkg p ->
				      "-add " ^ 
					Godi_file.uquote 
					(Godi_file.slashname p#binary_archive)
				  | _ -> assert false
			     )
			     (List.rev !to_do) in
			 let exe =
			   match Sys.os_type with
			     | "Win32" | "Cygwin" -> ".exe"
				 (* Required for bootstrap because we have
                                    temporarily both a shell script and an
                                    .exe
				  *)
			     | _ -> "" in
			 let restart_binary =
			   Godi_file.addname
			     (Godi_file.filename (Godi_query.localbase()))
			     ("sbin/godi_console" ^ exe) in
			 let restart_cmd =
			   "exec " ^ 
			     (Godi_file.slashname restart_binary) ^ 
			     " perform" ^ 
			     " -tmpdir-state '%s' " ^ 
			     String.concat " " restart_args in
			 raise(Restart(sh_cmd ^ "\n" ^ restart_cmd));
			 
		     | User_error e when !as_much_as_possible ->
			 printf "### Error: %s\n%!" e;
			 let to_disable =
			   Godi_plan.get_dependent_packages
			     ~include_build_deps:true
			     ~active:to_build_or_add'
			     ~selected:[pkg]
			     () in
			 Hashtbl.replace disabled_pkg (pkg_name pkg) ();
			 List.iter
			   (fun p ->
			      let pn = pkg_name p in
			      printf "# Disabling dependent package: %s\n%!" pn;
			      Hashtbl.replace disabled_pkg pn ()
			   )
			   to_disable
		 )
	       )
	   | `Binary ->
	       let n = pkg_name pkg in
	       if Hashtbl.mem disabled_pkg n then
		 printf "### Skipping add of %s because of build errors in previous packages\n" n
	       else (
		 printf "### Adding %s\n" (pkg_fullname pkg);
		 flush stdout;
		 if not !dry_run then (
		   Godi_perform.add_pkg con pkg;
		   (any_pkg pkg)#set_user_status `Keep;
		   save_wishes();
		 )
	       )
	   | `Installed | `Dispersed ->
	       assert false
      )
      to_build_or_add'
  );

  let have_error = ref false in
  Hashtbl.iter
    (fun n () ->
       printf "### Not built because of errors: %s\n" n;
       have_error := true;
    )
    disabled_pkg;
  flush stdout;
  
  if !have_error then
    raise(User_error "Some packages have not been built")
;;


let print_deps l =
  String.concat ", "
    (List.map Godi_versions.print_dependency l)
;;


let indent s =
  Str.global_replace (Str.regexp "\n") "\n    " s
;;


let remove_linefeeds s =
  Str.global_replace (Str.regexp "[\r\n]") " " s
;;


let version_str pkg =
  let v = pkg_version pkg in
  let r = pkg_pkgrevision pkg in
  if r > 0 then v ^ "godi" ^ string_of_int r else v
;;


let print_file_list indent pkg =
  let is = String.make indent ' ' in
  List.iter
    (function
       | `File(n,_) ->
	   printf "%s%s\n" is n
       | _ ->
	   ()
    )
    pkg # inst_plist
  


let long_description ?(file_list=false) pkg =
  (* Long format:
   * - The package name is on column 0
   * - The attributes begin on column 2
   * - Continuation lines begin on column 4
   * ==> Easy to parse!
   *)
  printf "%s\n" (pkg_name pkg);
  ( match pkgtype pkg with
	`Source -> 
	  printf "  type: source\n"
      | `Installed ->
	  printf "  type: installed\n"
      | `Binary ->
	  printf "  type: binary\n"
      | _ -> assert false
  );
  printf "  version: %s\n" (version_str pkg);
  printf "  comment: %s\n" (remove_linefeeds (pkg_comment pkg));
  printf "  description: %s\n" (indent (pkg_description pkg));
  printf "  dependencies: %s\n" (print_deps
				   (pkg_dependencies pkg));
  printf "  build_deps: %s\n" (print_deps
				 (pkg_build_dependencies pkg));
  ( match pkgtype pkg with
	`Source ->
	  printf "  build_message: %s\n"
	    (indent (source_pkg pkg)#build_message);
	  let nv = (source_pkg pkg)#values_of_build_options in
	  printf "  build_options: %s\n"
	    (indent
	       (String.concat "\n" 
		  (List.map
		     (fun (n,v) -> n ^ "=" ^ v)
		     nv)));
      | `Installed ->
	  let ipkg = installed_pkg pkg in
	  printf "  install message: %s\n" (indent ipkg#install_message);
          printf "  build_version_fingerprint: %s\n" 
            (indent ipkg#build_version_fingerprint);
	  printf "  build_version: %s\n" (indent ipkg#build_version);
	  printf "  build_info: %s\n" (indent ipkg#build_info);
	  ( match ipkg#requirements_script with
	      | None -> ()
	      | Some s -> printf "  requirements_script:\n    %s\n" (indent s)
	  );
	  ( match ipkg#install_script with
	      | None -> ()
	      | Some s -> printf "  install_script:\n    %s\n" (indent s)
	  );
	  ( match ipkg#deinstall_script with
	      | None -> ()
	      | Some s -> printf "  deinstall_script:\n    %s\n" (indent s)
	  );
	  printf "  preserved: %b\n"
	    (installed_pkg pkg)#preserve;
	  if file_list then (
	    printf "  files:\n";
	    print_file_list 4 ipkg
	  )
      | `Binary ->
	  let bpkg = binary_pkg pkg in
	  printf "  install message: %s\n" (indent bpkg#install_message);
	  printf "  build_version_fingerprint: %s\n" 
            (indent bpkg#build_version_fingerprint);
	  printf "  build_version: %s\n" (indent bpkg#build_version);
	  printf "  build_info: %s\n" (indent bpkg#build_info);
	  ( match bpkg#requirements_script with
	      | None -> ()
	      | Some s -> printf "  requirements_script:\n    %s\n" (indent s)
	  );
	  ( match bpkg#install_script with
	      | None -> ()
	      | Some s -> printf "  install_script:\n    %s\n" (indent s)
	  );
	  ( match bpkg#deinstall_script with
	      | None -> ()
	      | Some s -> printf "  deinstall_script:\n    %s\n" (indent s)
	  );
	  if file_list then (
	    printf "  files:\n";
	    print_file_list 4 bpkg
	  )
      | _ ->
	  ()
  );
  printf "\n"
;;



let cmd_list() =
  let con = stderr_console() in
  let list_src = ref false in
  let list_inst = ref false in
  let only_names = ref false in
  let long = ref false in
  let filters = ref [] in

  Arg.parse 
      [ "-source", Arg.Set list_src,
                "           List packages available as source code";
	"-installed", Arg.Set list_inst,
	           "        List installed packages";
	"-names", Arg.Set only_names,
	       "            Output the pure list of package names (no versions etc.)";
	"-long", Arg.Set long,
	      "             Output descriptions, options, etc.";
	
      ]
      (fun s -> 
	 let dep = Godi_versions.parse_dependency ~sloppy:true s in
	 filters := !filters @ [dep])
      (helptext "[<options>] [<pkgfilter> ...]");

  if not !list_src && not !list_inst then (
    list_src := true;
    list_inst := true;
  );

  (* First scan packages *)
  let src_pkgs = if !list_src || !long then scan_source_packages con else [] in
  let inst_pkgs = if !list_inst || !long then get_installed_packages con else [] in
  let out_pkgs = src_pkgs @ inst_pkgs in

  (* filters: Select which packages are printed *)
  let is_selected pkg =
    (!filters = []) ||
    ( (* Package is selected when there is a matching filter *)
      List.exists
	(fun (n,cond) ->
	   (n = (pkg_name pkg)) &&
	   (Godi_versions.version_matches_condition
	      (pkg_version pkg) cond))
	!filters
    )
  in

  (* Arrange packages by name: *)
  let h = Hashtbl.create 20 in
  let keys = ref [] in
  List.iter
    (fun pkg ->
       let n = pkg_name pkg in
       let old =
	 try Hashtbl.find h n with Not_found -> [] in
       if old = [] then keys := n :: !keys;
       Hashtbl.replace h n (pkg :: old)
    )
    out_pkgs;

  keys := List.sort compare !keys;

  (* Output list *)
  List.iter
    (fun name ->
       let pkgs = Hashtbl.find h name in
       let (inst_version, inst_sel) =
	 try
	   let inst_pkg = 
	     List.find (fun pkg -> pkgtype pkg = `Installed) pkgs in
	   ( Some(version_str inst_pkg),
	     !list_inst && is_selected inst_pkg
	   )
	 with
	     Not_found -> (None, false) in
       let (src_version, src_sel) =
	 try
	   let src_pkg = 
	     List.find (fun pkg -> pkgtype pkg = `Source) pkgs in
	   ( Some(version_str src_pkg),
	     !list_src && is_selected src_pkg
	   )
	 with
	     Not_found -> (None, false) in

       let sel = inst_sel || src_sel in
       
       if sel then (
	 if !only_names then
	   print_endline name
	 else (
	   let inst_version_str =
	     match inst_version with
		 None -> "-"
	       | Some v -> v in
	   let src_version_str =
	     match src_version with
		 None -> "-"
	       | Some v -> v in

	   if !long then (
	     List.iter
               (fun pkg ->			    
		  let sel =
		    match pkgtype pkg with
			`Source -> !list_src
		      | `Installed -> !list_inst
		      | _ -> false in
		  if sel then
		    long_description pkg
	       )
	       pkgs

	   )
	   else (
	     ( match !list_src, !list_inst with
		 | true, false ->
		     printf "%-30s src:%-10s\n"
		       name
		       src_version_str
		 | false, true ->
		     printf "%-30s inst:%-10s\n"
		       name
		       inst_version_str
		 | true, true ->
		     printf "%-30s inst:%-10s src:%-10s\n"
		       name
		       inst_version_str
		       src_version_str
		 | _ -> assert false
	     )
	   )
	 )
       )
    )
    !keys;
;;


let cmd_info() =
  let con = stderr_console() in
  let long = ref false in
  let file_list = ref false in
  let exprs = ref [] in

  Arg.parse 
      [ "-long", Arg.Set long,
	"      Output descriptions, options, etc.";

	"-file-list", Arg.Set file_list,
	"           Output file list";
      ]
      (fun s -> 
	 if not(String.contains s '/') then (
	   let dep = Godi_versions.parse_dependency ~sloppy:true s in
	   exprs := !exprs @ [ `Filter dep ]
	 )
	 else
	   exprs := !exprs @ [ `File s ]
      )
      (helptext "[<options>] [<pkgfilter> | <file> | <URL> ...]");

  (* First scan packages *)
  let inst_pkgs = get_installed_packages con in
  let inst_pkgs =  (* sort *)
    List.sort
      (fun p1 p2 -> compare (pkg_name p1) (pkg_name p2))
      inst_pkgs in
    
  if !exprs = [] then
    exprs := List.map (fun pkg -> `Pkg pkg) inst_pkgs;

  (* Now expand exprs: *)
  let pkg_list =
    List.flatten
      (List.map
	 (function
	    | `Pkg pkg -> [pkg]
	    | `Filter (nm,cond) ->
		List.filter
		  (fun pkg ->
		     (nm = (pkg_name pkg)) &&
		       (Godi_versions.version_matches_condition
			  (pkg_version pkg) cond))
		  inst_pkgs
	    | `File fn ->
		let real_fn, cleanup = local_or_remote_file con fn in
		[Godi_query.get_binary_package real_fn]
		  (* We ignore [cleanup] because all tmp dirs are
                     deleted anyway when the command finishes
                   *)
	 )
	 !exprs
      ) in

  let p_files i (pkg : package) =
    match pkgtype pkg with
      | `Binary -> print_file_list i (binary_pkg pkg)
      | `Installed -> print_file_list i (installed_pkg pkg)
      | _ -> ()
  in

  (* Output list *)
  let n = List.length pkg_list in
  List.iter
    (fun pkg ->
       if !long then
	 long_description ~file_list:!file_list pkg
       else (
	 if n = 1 && !file_list then
	   p_files 0 pkg
	 else (
	   printf "%-30s %s\n"
	     (pkg_fullname pkg)
	     (remove_linefeeds (pkg_comment pkg));
	   if !file_list then
	     p_files 4 pkg
	 )
       )
    )
    pkg_list
;;


let cmd_dep() =
  let con = stderr_console() in
  let mode = ref `Normal in
  let query_type = ref `Names in
  let dep_expressions = ref [] in

  Arg.parse
    [ "-quiet", Arg.Unit (fun () -> mode := `Quiet),
      "  Only output whether the dependencies are fulfilled but not how";
      
      "-verbose", Arg.Unit (fun () -> mode := `Verbose),
      "  Output the result in a verbose style";
      
      "-names", Arg.Unit (fun () -> query_type := `Names),
      "  Print the actually matching pkg names with version numbers for every dep";
      
      "-names-all", Arg.Unit (fun () -> query_type := `Names_all),
      "  Print the matching pkg names+versions for all deps including prerequisites";
      
      "-size", Arg.Unit (fun () -> query_type := `Size),
      "  Print the package size in bytes for every dep";
      
      "-size-all", Arg.Unit (fun () -> query_type := `Size_all),
      "  Print the summed size of the matching pkgs and all prerequisites in bytes";
    ]
    (fun s -> dep_expressions := !dep_expressions @ [s])
    (helptext "[<options>] <pgk-dep> ...");

  let inst_pkgs = Godi_query.get_installed_packages con in
  (* Read packages now! *)

  let success = ref true in

  let pkg_opt_list =
    (* The list of installed packages corresponding to dep_expressions *)
    List.map
      (fun dep ->
	 let (name, cond) = 
	   try Godi_versions.parse_dependency dep
	   with Failure msg -> raise (User_error msg)
	 in

	 try
	   let pkg = Godi_query.get_this_installed_package con name in 
	   (* or Not_found *)
	   if Godi_query.matches_dep pkg (name, cond) then
	     (dep, Some pkg)
	   else
	     (dep, None)
	 with
	     Not_found -> (dep, None)
      )
      !dep_expressions in

  (match !query_type with
     | `Names
     | `Size ->
	 (* Process every dependency separately! *)
	 List.iter
	   (fun (dep, pkg_opt) ->
	      success := !success && (pkg_opt <> None);
	      match !mode with
		| `Quiet -> ()
		| `Normal ->
		    ( match pkg_opt with
			| None -> 
			    eprintf "%s: not found\n" dep; flush stderr
			| Some pkg ->
			    ( match !query_type with
				| `Names ->
				    printf "%s\n" (pkg_fullname pkg); 
				    flush stdout
				| `Size ->
				    let ipkg = installed_pkg pkg in
				    printf "%Ld\n" (ipkg # size); flush stdout
				| _ -> assert false
			    )
		    )
		| `Verbose ->
		    ( match pkg_opt with
			| None -> 
			    eprintf "%s: not found\n" dep; flush stderr
			| Some pkg ->
			    let n = pkg_fullname pkg in
			    ( match !query_type with
				| `Names ->
				    printf "%s: %s\n" dep n; flush stdout
				| `Size ->
				    let ipkg = installed_pkg pkg in
				    printf "%s: %Ld\n" n (ipkg # size); 
				    flush stdout
				| _ -> assert false
			    )
		    )
	   )
	   pkg_opt_list

     | `Names_all
     | `Size_all ->
	 (* First compute the closure of the set of packages: *)
	 let pkgs =
	   List.flatten
	     (List.map
		(fun (dep, pkg_opt) ->
		   success := !success && (pkg_opt <> None);
		   match pkg_opt with
		     | None ->
			 eprintf "%s: not found\n" dep; flush stderr;
			 []
		     | Some pkg ->
			 [pkg]
		)
		pkg_opt_list 
	     ) in

	 let reqs =
	   Godi_plan.get_required_packages
	     ~active:inst_pkgs
	     ~selected:pkgs
	     () in

	 let all = 
	   Godi_plan.sort_packages (pkgs @ reqs) in

	 (* all: The packages mentioned in the dep expressions plus all
          * prerequisites, sorted topologically.
          *)

	 let total_size() =
	   List.fold_left
	     (fun acc pkg -> Int64.add acc ((installed_pkg pkg) # size))
	     0L
	     all in

	 match !mode with
	   | `Quiet -> ()
	   | `Normal ->
	       ( match !query_type with
		   | `Names_all ->
		       List.iter
			 (fun pkg ->
			    printf "%s\n" (pkg_fullname pkg); 
			    flush stdout
			 )
			 all
		   | `Size_all ->
		       printf "%Ld\n" (total_size());
		       flush stdout
		   | _ ->
		       assert false
	       )
	       
	   | `Verbose ->
	       ( match !query_type with
		   | `Names_all ->
		       List.iter
			 (fun pkg ->
			    printf "%s\n" (pkg_fullname pkg); 
			    flush stdout
			      (* Don't know how to be verbose here! *)
			 )
			 all
		   | `Size_all ->
		       printf "Total size: %Ld\n" (total_size());
		       flush stdout
		   | _ ->
		       assert false
	       )
  );

  if not !success then
    raise (Exit_code 1)
;;


let cmd_delete() =
  let con = stderr_console() in
  let mode = ref `Normal in
  let recurse = ref false in
  let verbose = ref false in
  let package_names = ref [] in

  Arg.parse
      [ "-f", Arg.Unit (fun () -> 
			  if !mode <> `Normal then 
			    raise(Arg.Bad "Incompatible switches");
			  mode := `Force),
	   "   Force deletion";
	"-n", Arg.Unit (fun () -> 
			  if !mode <> `Normal then 
			    raise(Arg.Bad "Incompatible switches");
			  mode := `Simulate),
	   "   Only simulate deletion";
	"-r", Arg.Set recurse,
	   "   Delete successor packages";
	"-v", Arg.Set verbose,
	   "   Turn on verbose output";
      ]
      (fun s -> package_names := !package_names @ [s])
      "usage: godi_console delete [<options>] <pkg>...";

  (* -f and -r are incompatible: Too risky *)
  if !mode = `Force && !recurse then
    user_error("The options -f and -r are incompatible with each other");

  (* General algorithm:
   *
   * The listed packages are looked up, and the package objects are passed
   * to the planner. The result is that the planner suggests to delete all
   * successor packages, and the listed packages. If -r, we can just do that.
   * If not -r, and there are packages not listed on the command line, stop
   * with an error message.
   *
   * The option -f has two effects: First, it is now allowed that the planner
   * returns successor packages. These are not deleted, however, but just
   * ignored. The [mode] argument is passed to [remove_pkg], so the following
   * is turned off:
   * - Checksum test
   * - Requirements script
   *)
  let message = printmsg con in
  let inst_pkgs = get_installed_packages ~verbose:!verbose con in

  let sel_pkgs =
    List.map
      (fun s ->
	 (* If the package name has a version string, we accept that. It is
	  * then tested that the version matches.
	  *)
	 inst_pkg_by_name con s
      )
      !package_names in
  List.iter
    (fun pkg -> (installed_pkg pkg) # set_user_status `Remove)
    sel_pkgs;
  let _sel_pkgs_set = PkgMap.make (List.map (fun p -> (p,())) sel_pkgs) in

  if !verbose then
    message "Planning the removal\n";
  let n = Godi_plan.update_plan 
	    ~iteration:1 
	    inst_pkgs in
  (* May raise Mission_impossible, or Package_installation_not_unique *)

  if n>0 && not !recurse && !mode <> `Force then
    user_error("There are dependent packages, cannot delete");

  let to_remove =
    List.filter to_be_removed inst_pkgs in

  let to_remove' =
    List.rev(Godi_plan.sort_packages to_remove) in

  let had_errors = ref false in

  if to_remove = [] then
    ( if !verbose then 
	message "Nothing to do.\n" )
  else (
    List.iter
	  (fun pkg ->
	     if (installed_pkg pkg)#preserve && !mode <> `Force then
	       message
	         (sprintf "Package %s cannot be removed, skipping\n" (pkg_fullname pkg))
	     else (
	       let do_remove =
		 match pkg_user_status pkg with
		     `Remove -> true
		   | `Remove_dep _ -> !recurse
		   | _ -> false 
	       in
	       if do_remove then (
		 if !verbose then
		   message
		     (sprintf "Removing %s\n" (pkg_fullname pkg));
		 ( try
		     Godi_perform.remove_pkg 
		       ~soft_errors:true ~mode:!mode con pkg;
		   with
		       Godi_perform.Soft_rm_error ->
			 had_errors := true
		 )
	       )
	     )
	  )
	  to_remove';
  );
  if !had_errors then
    raise(Exit_code 1)
;;


let cmd_add() =
  let con = stderr_console() in
  let mode = ref `Normal in
  let upgrade = ref false in
  let verbose = ref false in
  let url_opt = ref None in
  let package_selectors = ref [] in

  let parse_selector s =
    if ( Filename.is_implicit s &&
           not(Godi_file_handler.is_url s) &&
           not (Filename.check_suffix s ".tgz")
       )
    then
      `Dep(Godi_versions.parse_dependency ~sloppy:true s)
    else
      `File(fst(local_or_remote_file con s))
  in

  Arg.parse
      [ "-u", Arg.Set upgrade,
	   "   Replace installed package";
	"-n", Arg.Unit (fun () -> mode := `Simulate),
	   "   Only simulate addition";
        "-d", Arg.String (fun s -> url_opt := Some s),
        "<url>  Take packages from this URL (or directory), \
         overriding GODI_BINPKG";
	"-v", Arg.Set verbose,
	   "   Turn on verbose output";
	(* Later:
	 * -pool <location>: Take package files from this pool. Also
	 *    GODI_BINARY_POOLS environment variable for the same purpose.
	 * -r: Add missing requirements automatically
	 *
	 * Idea:
	 * -f: We can force the deletion (in conjuction with -u)
	 *)
      ]
      (fun s -> 
	 package_selectors := 
           !package_selectors @ [parse_selector s])
    (helptext "[<options>] <pkgfile.tgz> or <URL> or <dep>...");

  let message = printmsg con in
  let inst_pkgs = get_installed_packages ~verbose:!verbose con in
  
  let dep_list_raw, filenames_raw =
    List.partition
      (function
         | `Dep _ -> true
         | `File _ -> false
      )
      !package_selectors in

  let dep_list =
    List.map
      (function
         | `Dep dep -> dep
         | _ -> assert false
      )
      dep_list_raw in

  let filenames =
    List.map
      (function
         | `File name -> name
         | _ -> assert false
      )
      filenames_raw in

  let all_pkgs =
    if dep_list <> [] then (
      let l =
        match !url_opt with
          | Some url ->
              if !verbose then
                message ("Scanning packages at " ^ url ^ "...\n");
              Godi_query.scan_binary_packages con url
          | None ->
              if !verbose then
                message "Scanning package sources in GODI_BINPKG...\n";
              Godi_query.scan_all_binary_packages con
      in
      if !verbose then (
        message "Found packages:\n";
        List.iter (fun p -> message ("  " ^ pkg_fullname_at p ^ 
                                       "\n")) l
      );
      l
    )
    else [] in

  if !verbose && filenames <> [] then
    message "Scanning files to install...\n";
  
  let binary_pkgs =
    List.map
      (fun name ->
         get_binary_package name
      )
      filenames in
  
  List.iter
    (fun pkg -> (binary_pkg pkg) # set_user_status `Add)
    binary_pkgs;

  if !verbose then
    message "Planning the addition\n";
  let _n = Godi_plan.update_plan 
            ~select_by_dep:dep_list
	    ~iteration:1 
	    (inst_pkgs @ all_pkgs @ binary_pkgs) in
  (* May raise Mission_impossible, or Package_installation_not_unique *)

  let to_remove =
    List.filter to_be_removed inst_pkgs in

  let to_remove' =
    List.rev(Godi_plan.sort_packages to_remove) in
 
  let to_add =
    List.filter to_be_added (all_pkgs @ binary_pkgs) in

  let to_add' =
    Godi_plan.sort_packages to_add in

  if to_remove <> [] && not !upgrade then
    user_error("Some packages are already installed, cannot add");

  (* Execute the plan: *)

  let had_errors = ref false in

  if to_remove = [] && to_add = [] then
    ( if !verbose then 
	message "Nothing to do.\n" )
  else (
    List.iter
	  (fun pkg ->
	     if (installed_pkg pkg)#preserve then
	       message
	         (sprintf "Package %s cannot be removed, skipping\n" (pkg_fullname pkg))
	     else (
	       if !verbose then
		 message
		   (sprintf "Removing %s\n" (pkg_fullname pkg));
	       ( try
		   Godi_perform.remove_pkg 
		     ~soft_errors:true
		     ~mode:(!mode :> Godi_perform.rm_mode) con pkg;
		 with
		     Godi_perform.Soft_rm_error ->
		       had_errors := true
	       )
	     )
	  )
	  to_remove';

    List.iter
      (fun pkg ->
	 if !verbose then
	   message (sprintf "Adding %s\n" (pkg_fullname_at pkg));
	 Godi_perform.add_pkg ~mode:!mode con pkg;
      )
      to_add'
  );
  if !had_errors then
    raise(Exit_code 1)
;;



let colon_re = Str.regexp ":.*$";;


let cmd_ptool() =
  let con = stderr_console() in
  let input = ref None in
  let output = ref None in

  let prefix = ref None in
  let dispersion = ref None in

  let fixup_symlinks = ref false in

  let disp = ref false in
  let disp_name = ref None in
  let disp_comment = ref None in
  let disp_description = ref None in
  let disp_maintainer = ref "" in
  let disp_dependencies = ref [] in
  let disp_build_dependencies = ref [] in
  let disp_conflicts = ref [] in
  let disp_install_message = ref "" in
  let disp_build_version = ref "" in
  let disp_build_info = ref "" in
  let disp_preserve = ref false in
  let disp_req_script = ref None in
  let disp_install_script = ref None in
  let disp_deinstall_script = ref None in
  let disp_src_fp = ref None in

  let print_out_name = ref false in

  let parse_dep_list s =
    let l = split s in
    List.map (fun  s -> 
		(* Strip ":"-part off *)
		let s' = Str.global_replace colon_re "" s in
		Godi_versions.parse_dependency s') l
  in

  let only_one_input() =
    if !input <> None then
      user_error "It is only possible to have one input"
  in

  let only_one_output() =
    if !output <> None then
      user_error "It is only possible to have one output"
  in


  Arg.parse
      [ "-in-installed",
	Arg.String (fun s -> only_one_input(); input := Some(`Installed s)),
	"<pkg-name>  Process this installed package";

	"-in-dispersed",
	Arg.Unit (fun () -> only_one_input(); input := Some `Dispersed),
	"  Process dispersed files characterized by -set-* options";

        "-in-binary",
        Arg.String (fun s -> only_one_input(); input := Some(`Binary s)),
        "<file.tgz>  Process this binary package";

	"-fixup-symlinks",
	Arg.Set fixup_symlinks,
	"  Modify the input, and fix up non-relative symlinks";

        "-out-file-list",
        Arg.Unit (fun () -> only_one_output(); output := Some `File_list),
        "  Print the list of regular files to stdout";

        "-out-manifest",
        Arg.Unit (fun () -> only_one_output(); output := Some `Manifest),
        "  Print a PKGMANIFEST fragment for this package to stdout";

	"-out-register",
	Arg.Unit (fun () -> only_one_output(); output := Some `Register),
	"  Register the input in the GODI package database";

	"-out-install",
	Arg.Unit (fun () -> only_one_output(); output := Some `Install),
	"  Install the input as GODI package (scripts are not run)";

	"-out-binary",
	Arg.String (fun s -> only_one_output(); output := Some (`Binary s)),
	"<dir>  Create a binary package in the directory <dir>";

        "-print-out-name",
        Arg.Set print_out_name,
        "   Print the name of the binary package to stdout";

	"-set-prefix",
	Arg.String (fun s -> disp := true; prefix := Some s),
	"<prefix>  Take dispersed files from this directory hierarchy";

	"-set-plist",
	Arg.String (fun s -> disp := true; dispersion := Some(`Plist s)),
	"<file>  Select dispersed files matching this PLIST <file>";

	"-set-auto-plist",
	Arg.Unit (fun () -> disp := true; dispersion := Some `Auto_plist),
	"  Select all files in <prefix> (autogenerated PLIST)";
	
	"-set-name",
	Arg.String (fun s -> disp := true; disp_name := Some s),
	"<pkg-name>  Set the name of the package";

	"-set-comment",
	Arg.String (fun s -> disp := true; disp_comment := Some s),
	"<text>  Set the comment string of the package";

	"-set-description",
	Arg.String (fun s -> disp := true; disp_description := Some s),
	"<text>  Set the description text of the package";

	"-set-description-file",
	Arg.String (fun s -> 
		      let fn = Godi_file.filename s in
		      disp := true; 
		      disp_description := 
			Some(read_text_file ~pref_user_error:true fn)),
	"<file>  Take the description text of the package from <file>";

	"-set-maintainer",
	Arg.String (fun s -> disp := true; disp_maintainer := s),
	"<text>  Set the name of the GODI maintainer";

	"-set-dependencies",
	Arg.String (fun s -> 
		      disp := true; disp_dependencies := parse_dep_list s),
	"<deps>  Set the (space-seperated) dependencies";

	"-set-build-dependencies",
	Arg.String (fun s -> 
		      disp := true; 
		      disp_build_dependencies := parse_dep_list s),
	"<deps>  Set the (space-seperated) build dependencies";

	"-set-conflicts",
	Arg.String (fun s -> disp := true; disp_conflicts := parse_dep_list s),
	"<deps>  Set the (space-seperated) conflicts";

	"-set-src-fingerprint",
	Arg.String (fun s -> disp := true; disp_src_fp := Some s),
	"<fp>   Set the source fingerprint";

	"-set-install-message",
	Arg.String (fun s -> disp := true; disp_install_message := s),
	"<text>  Set the message printed on installation";

	"-set-install-message-file",
	Arg.String (fun s -> 
		      let fn = Godi_file.filename s in
		      disp := true; 
		      disp_install_message := 
			(read_text_file ~pref_user_error:true fn)),
	"<file>  Take the message printed on installation from <file>";

	"-set-build-version",
	Arg.String (fun s -> disp := true; disp_build_version := s),
	"<text>  Set the build version text";

	"-set-build-version-file",
	Arg.String (fun s -> 
		      let fn = Godi_file.filename s in
		      disp := true; 
		      disp_build_version := 
			(read_text_file ~pref_user_error:true fn)),
	"<file>  Take the build version text from <file>";

	"-set-build-info",
	Arg.String (fun s -> disp := true; disp_build_info := s),
	"<text>  Set the build info text";

	"-set-build-info-file",
	Arg.String (fun s -> 
		      let fn = Godi_file.filename s in
		      disp := true; 
		      disp_build_info := 
			(read_text_file ~pref_user_error:true fn)),
	"<file>  Take the build info text from <file>";

	"-set-preserve",
	Arg.Bool (fun b -> disp := true; disp_preserve := b),
	"<bool>  Set whether the package is protected from being deleted";

	"-set-req-script",
	Arg.String (fun s -> disp := true; disp_req_script := Some s),
	"<commands>  Set the requirements script";

	"-set-req-script-file",
	Arg.String (fun s -> 
		      let fn = Godi_file.filename s in
		      disp := true;
		      disp_req_script := 
			Some (read_text_file ~pref_user_error:true fn)),
	"<file>   Set the requirements script from a file";

	"-set-install-script",
	Arg.String (fun s -> disp := true; disp_install_script := Some s),
	"<commands>  Set the install script";

	"-set-install-script-file",
	Arg.String (fun s -> 
		      let fn = Godi_file.filename s in
		      disp := true;
		      disp_install_script := 
			Some (read_text_file ~pref_user_error:true fn)),
	"<file>   Set the install script from a file";

	"-set-deinstall-script",
	Arg.String (fun s -> disp := true; disp_deinstall_script := Some s),
	"<commands>  Set the deinstall script";

	"-set-deinstall-script-file",
	Arg.String (fun s -> 
		      let fn = Godi_file.filename s in
		      disp := true;
		      disp_deinstall_script := 
			Some (read_text_file ~pref_user_error:true fn)),
	"<file>   Set the deinstall script from a file";

      ]
    (fun s -> raise(Arg.Bad "Unexpected argument"))
    (helptext "<option>...");

  if !input = None then
    user_error "Missing input";

  if !output = None then
    user_error "Missing output";

  if !input <> Some `Dispersed && !disp then
    user_error "The -set-* options are only allowed for -in-dispersed";

  ( match !input with
      | Some (`Binary _) ->
          ( match !output with
              | Some `Register ->
                  user_error "A binary package cannot be registered"
              | Some `Install ->
                  user_error "A binary package cannot be installed with \
                              this tool; use godi_add"
              | _ ->
                  ()
          )
      | _ -> ()
  );

  let in_pkg =
    match !input with
      | None -> user_error "Missing input"
      | Some (`Installed pkg_name) ->
	  let pkg = inst_pkg_by_name con pkg_name in
	  Godi_ptool.dispersed_of_installed_package con pkg
      | Some (`Binary pkg_file) ->
          Godi_query.get_binary_package (Godi_file.filename pkg_file)
      | Some `Dispersed ->
	  let need_arg var name =
	    match !var with
	      | None -> user_error ("Missing parameter: " ^ name)
	      | Some v -> v
	  in
	  let full_name = need_arg disp_name "-set-name" in
	  let (name, version, pkgrevision) = 
	    Godi_versions.split_name_version full_name in
	  
	  let prefix = Godi_file.filename(need_arg prefix "-set-prefix") in

	  let d =
	    match !dispersion with
	      | None ->
		  user_error "Need either -set-plist or -set-auto-plist"
	      | Some (`Plist f) ->
		  let plist = 
		    Godi_plist.parse_user_plist (Godi_file.filename f) in
		  Godi_ptool.user_plist_dispersion ~root:prefix plist
	      | Some `Auto_plist ->
		  Godi_ptool.stage_dispersion prefix
	  in

	  Godi_ptool.create_dispersed_package
	    ~name ~version ~pkgrevision
	    ~comment:(need_arg disp_comment "-set-comment")
	    ~description:(need_arg disp_description "-set-description")
	    ~godi_maintainer:!disp_maintainer
	    ~dependencies:!disp_dependencies
	    ~build_dependencies:!disp_build_dependencies
	    ~conflicts:!disp_conflicts
	    ~install_message:!disp_install_message
	    ~build_version:!disp_build_version
	    ~build_info:!disp_build_info
	    ~src_fingerprint:!disp_src_fp
	    ~preserve:!disp_preserve
	    ~requirements_script:!disp_req_script
	    ~install_script:!disp_install_script
	    ~deinstall_script:!disp_deinstall_script
	    d con
  in

  if !fixup_symlinks then
    Godi_ptool.fixup_symlinks con (dispersed_pkg in_pkg)#dispersion;

  match !output with
    | None -> assert false
    | Some `Register ->
	Godi_perform.register_pkg con in_pkg
    | Some `Install ->
	Godi_perform.install_pkg con in_pkg
    | Some (`Binary dir) ->
	let out_file_name =
	  Godi_file.addsuffix
	    (Godi_file.addname
	       (Godi_file.filename dir)
               (Godi_query.binary_file_base in_pkg)
	    )
	    ".tgz" in
	let out_pkg =
	  Godi_ptool.binary_of_dispersed_package con out_file_name in_pkg in
        let n = Godi_file.slashname (binary_pkg out_pkg)#binary_archive in
	eprintf "Wrote %s\n" n;
        if !print_out_name then
          print_endline n
    | Some `File_list ->
        let plist =
          match in_pkg with
            | `Binary_pkg p -> p # inst_plist
            | `Installed_pkg p -> p # inst_plist
            | `Dispersed_pkg p -> p # inst_plist
            | `Wrapped_pkg _ -> []
            | `Source_pkg _ -> [] in
        List.iter
          (function
             | `File(name,_) ->
                 print_endline name
             | _ ->
                 ()
          )
          plist
    | Some `Manifest ->
        let plist =
          match in_pkg with
            | `Binary_pkg p -> p # inst_plist
            | `Installed_pkg p -> p # inst_plist
            | `Dispersed_pkg p -> p # inst_plist
            | `Wrapped_pkg _ 
            | `Source_pkg _ -> assert false in
        let plist' =
          List.filter
            (function
               | #Godi_plist.core_plist_element -> false
               | _ -> true
            )
            plist in
        let plist'' =
          plist' @ [ `Comment (pkg_comment in_pkg);
                     `Descr (pkg_description in_pkg)
                   ] in
        let have_fp =
          List.exists
            (function `Bldversfingerprint _ -> true|_ -> false) 
            plist'' in
        let plist''' =
          if have_fp then
            plist''
          else
            plist'' @
              [ `Bldversfingerprint
                  (  match in_pkg with
                       | `Binary_pkg p -> p # build_version_fingerprint
                       | `Installed_pkg p -> p # build_version_fingerprint
                       | `Dispersed_pkg p -> p # build_version_fingerprint
                       | _ -> assert false
                  )
              ] in
        let ch = new Netchannels.output_channel stdout in
        Godi_plist.print_inst_plist
          ch
          plist''';
        ch # output_string "\n";
        ch # flush()
;;
	

let cmd_initool() =
  let open Godi_pkglist in
  (* let con = stderr_console() in *)

  let input = ref [] in
  let output = ref None in

  let name_filter = ref None in
  let version_filter = ref None in
  let newest = ref false in
  let rocketboost = ref false in

  let mode = ref None in
  let pkg_source = ref "" in
  let pkg_name = ref "" in
  let pkg_version = ref "" in
  let pkg_comment = ref "" in
  let pkg_descr = ref "" in
  let pkg_maintainer = ref "" in
  let pkg_homepage = ref "" in
  let pkg_svn_rev = ref (Some 0) in
  let pkg_svn_path = ref "" in
  let pkg_date = Unix.gettimeofday() in
  let pkg_rel_by = ref (current_user()) in
  let pkg_section = ref "" in
  let pkg_distfiles = ref [] in

  Arg.parse
    [ "-o", Arg.String (fun s -> output := Some s),
      "<file>  The output is written to this file";

      "-newest", Arg.Set newest,
      "   Only output the newest definition (by timestamp) for a package name";

      "-rocketboost", Arg.Set rocketboost,
      "   Output in rocketboost-compatible 'available.new' format";

      "-filter", Arg.String (fun s ->
                               let (n,v) = Godi_versions.split_name_vsuffix s in
                               name_filter := Some n;
                               version_filter := 
                                 if v <> "" then Some v else None
                            ),
      "<namespec>  Only keep the package(s) with this name, or name-version";

      "-add", Arg.Unit (fun () -> mode := Some `Add),
      "   Add the package described by the -pkg-* options";

      "-delete", Arg.Unit (fun () -> mode := Some `Delete),
      "   Delete the package described by the -pkg-name option";

      "-pkg-name", Arg.String (fun s ->
                                 let (n,v) = 
                                   Godi_versions.split_name_vsuffix s in
                                 if v = "" then
                                   raise
                                     (Arg.Bad("-pkg-name: version required"));
                                 pkg_name := n;
                                 pkg_version := v
                              ),
      "<name-version>  The name and version of the package";

      "-pkg-source", Arg.Set_string pkg_source,
      "<source>   The source specifier of the package";

      "-pkg-comment", Arg.Set_string pkg_comment,
      "<text>   The comment string of the package";

      "-pkg-descr", Arg.Set_string pkg_descr,
      "<text>   The description string of the package (can be multi-line)";

      "-pkg-descr-file", Arg.String(fun s ->
                                      let f = Godi_file.filename s in
                                      pkg_descr :=
                                        Godi_file.read_file ~mode:`Text f),
      "<file>   The description string of the package, given via file";

      "-pkg-maintainer", Arg.Set_string pkg_maintainer,
      "<text>   The maintainer string of the package";

      "-pkg-homepage", Arg.Set_string pkg_homepage,
      "<text>   The homepage string of the package";

      "-pkg-svn-path", Arg.Set_string pkg_svn_path,
      "<text>   The svn path of the package (optional)";

      "-pkg-svn-rev", Arg.Int (fun n -> pkg_svn_rev := Some n),
      "<n>   The (integer) svn revision of the package (optional)";

      "-pkg-rel-by", Arg.Set_string pkg_rel_by,
      "<text>   By whom the package is released";

      "-pkg-section", Arg.Set_string pkg_section,
      "<text>   The GODI section";

      "-pkg-distfile", Arg.String (fun s -> 
                                     pkg_distfiles := 
                                       (Filename.basename s) :: !pkg_distfiles
                                  ),
      "<filename>   Add this distfile (-pkg-distfiles can be given multiple \
       times";
    ]
    (fun s ->
       input := Godi_file.filename s :: !input
    )
    (helptext " [<option>...] <package.ini> ...");

  if !mode <> None && (!pkg_name = "" || !pkg_version = "") then
    user_error ("For -add/-delete, the option -pkg-name is mandatory");

  if !mode = Some `Add then (
    if !pkg_source = "" then
      user_error "For -add, the source needs to be specified with -pkg-source";
    if !pkg_section = "" then
      user_error "For -add, the section needs to be specified with -pkg-section"
  );

  let old_pkgs =
    List.map
      (fun f ->
         Godi_pkglist.read_packages_ini (Godi_file.nativename f)
      )
      !input in

  if !mode = Some `Delete then (
    pkg_svn_rev := None;
    try
      let del_pkg = 
        List.find
          (fun pkg ->
             pkg.pkg_name = !pkg_name && pkg.pkg_version = !pkg_version
          )
          (List.flatten old_pkgs) in
      pkg_source := 
        (if !pkg_source = "" then del_pkg.pkg_source else !pkg_source);
      pkg_section := 
        (if !pkg_section = "" then del_pkg.pkg_section else !pkg_section);
    with
      | Not_found ->
          ()
  );

  let new_pkgs =
    if !mode = None then
      []
    else
      [ [ { pkg_source = !pkg_source;
            pkg_name = !pkg_name;
            pkg_version = !pkg_version;
            pkg_comment = !pkg_comment;
            pkg_descr = !pkg_descr;
            pkg_maintainer = !pkg_maintainer;
            pkg_homepage = !pkg_homepage;
            pkg_svn_path = !pkg_svn_path;
            pkg_svn_rev = !pkg_svn_rev;
            pkg_date = pkg_date;
            pkg_rel_by = !pkg_rel_by;
            pkg_section = !pkg_section;
            pkg_distfiles = List.rev !pkg_distfiles;
          }
        ]
      ] in
  
  let merged =
    Godi_pkglist.merge_packages_ini
      ~keep_only_newest:!newest
      (List.rev (new_pkgs @ old_pkgs)) in
  
  let filtered =
    List.filter
      (fun pkg ->
         (match !name_filter with
            | None -> true
            | Some n -> n = pkg.pkg_name
         ) &&
           ( match !version_filter with
               | None -> true
               | Some v -> v = pkg.pkg_version
           )
      )
      merged in

  let ch =
    match !output with
      | None -> stdout
      | Some file ->
          open_out (Godi_file.nativename (Godi_file.filename file)) in

  if !rocketboost then
    Godi_pkglist.write_available_new_ch ch filtered
  else
    Godi_pkglist.write_packages_ini_ch ch filtered;

  if !output <> None then
    close_out ch


let cmd_make argi =
  (* "make" is special, because we need to look at $MAKEFLAGS for additional
   * arguments. argi is the argument index with the first real argument
   * (after the command name).
   *)
  assert(argi=1 || argi=2);

  let args_makeflags =
    try
      let makeflags = Sys.getenv "MAKEFLAGS" in
      Array.of_list(Godi_match.parse_shell_words makeflags)
    with
	Not_found -> [| |] in
  let args_godi_makeflags =
    try
      let makeflags = Sys.getenv "GODI_MAKEFLAGS" in
      Array.of_list(Godi_match.parse_shell_words makeflags)
    with
	Not_found -> [| |] in
  let args_argv =
    (* Without command and "make" *)
    Array.sub Sys.argv argi (Array.length Sys.argv - argi) in
  let args =
    Array.concat [ [| "godi_make" |];    (* do this always! *)
		   args_godi_makeflags; 
		   args_makeflags; 
		   args_argv ] in
  (* Because later arguments override earlier ones, it is quite important
   * that [args_argv] is evaluated after [args_makeflags], so directly
   * passed arguments have precedence.
   *)

  let level =
    try
      int_of_string (Sys.getenv "MAKELEVEL")
    with
      | Failure _
      | Not_found -> 1
  in

  let makefile = ref None in

  (* This is portable *)
  if Sys.file_exists "makefile" then
    makefile := Some "makefile"
  else
    if Sys.file_exists "Makefile" then
      makefile := Some "Makefile";

  let inc_dirs = ref [ Godi_file.filename "." ] in
  let sys_inc_dirs = ref [] in
  let global_ignore_flag = ref false in
  let global_silent_flag = ref false in
  let exec_mode = ref `Execute in
  let mode = ref `Default in
  let vars = ref [] in
  let override = ref [] in
  let targets = ref [] in
  let sys_mk = ref true in
  let options = ref [] in
  let chdir = ref None in
  let trace_flag = ref false in
  let trace_script_flag = ref false in
  let trace_def_flag = ref false in

  Arg.current := 0;
  Arg.parse_argv
    args
    [ "-f", Arg.String (fun s -> makefile := Some s),
         "<file>  Set the Makefile to read";
      "-C", Arg.String (fun s -> chdir := Some s),
         "<dir>   Change into directory";
      "-I", Arg.String (fun s -> inc_dirs := 
			           !inc_dirs @ [Godi_file.filename s];
			         options := !options @ ["-I"; s]
		       ),
         "<dir>   Search in this directory";
      "-m", Arg.String (fun s -> sys_inc_dirs := 
			            !sys_inc_dirs @ [Godi_file.filename s];
			         options := !options @ ["-m"; s]
		       ),
         "<dir>   Search in this system directory";
      "-i", Arg.Unit (fun () -> global_ignore_flag := true;
			        options := !options @ ["-i"]
		     ),
         "        Ignore failed commands";
      "-n", Arg.Unit (fun () -> exec_mode := `Simulate_except;
			        options := !options @ ["-n"]
		     ),
         "        Simulate execution (except .MAKE targets)";
      "-N", Arg.Unit (fun () -> exec_mode := `Simulate;
			        options := !options @ ["-N"]
		     ),
         "        Simulate execution";
      "-q", Arg.Unit (fun () -> exec_mode := `Test;
			        options := !options @ ["-q"]
		     ),
         "        Only test whether targets are up to date";
      "-r", Arg.Unit (fun () -> sys_mk := false;
			        options := !options @ ["-r"]
		     ),
	 "        Do not include <sys.mk>";
      "-s", Arg.Unit (fun () -> global_silent_flag := true;
			        options := !options @ ["-s"]
		     ),
         "        Be silent";
      "-t", Arg.Unit (fun () -> exec_mode := `Fake;
			        options := !options @ ["-t"]
		     ),
         "        Fake the update of targets";
      "-trace", Arg.Unit (fun () -> trace_flag := true;
			            options := !options @ [ "-trace" ]),
             "    Trace the execution";
      "-trace-script", Arg.Unit (fun () -> trace_script_flag := true;
			            options := !options @ [ "-trace-script" ]),
             "    Call the shell with -x";
      "-trace-def", Arg.Unit (fun () -> trace_def_flag := true;
			            options := !options @ [ "-trace-def" ]),
             "    Trace the definitions";
      "-dump", Arg.Unit (fun () -> mode := `Dump),
            "     Create a dump file from the Makefile";
      "-V", Arg.String (fun s -> 
			  if !mode <> `Default && !mode <> `Variable then
			    raise(Arg.Bad "-V incompatible with specifying a target");
			  mode := `Variable;
			  vars := !vars @ [s]),
         "<var>   Output the unexpanded definition of <var>";
    ]
    (fun arg ->
       match decompose_assignment arg with
         | Some(lhs,rhs) ->
	     override := !override @ [ lhs, rhs ]
         | None ->
	     if !mode <> `Default && !mode <> `Target then
	       raise(Arg.Bad "-V incompatible with specifying a target");
	     mode := `Target;
	     targets := !targets @ [ arg ];
    )
    (if argi=1 then
       "usage: " ^ (Filename.basename Sys.argv.(0)) ^ 
	 " <options> ( var=value | target ) ..."
     else
       "usage: " ^ (Filename.basename Sys.argv.(0)) ^ " " ^ Sys.argv.(1) ^ 
	 " <options> ( var=value | target ) ..."
    );

  (match !chdir with
     | None -> ()
     | Some d -> 
	 Godi_file.chdir (Godi_file.filename d)
  );

  let ctx =
    Godi_make.create_context
      ~override:!override
      ~options:(Godi_match.quote_shell_words !options)
      ~user_mk_path:!inc_dirs
      ?sys_mk_path:(if !sys_inc_dirs = [] then None else Some !sys_inc_dirs)
      ~make_cmd:"godi_make"
      ~trace:!trace_flag
      ~trace_script:!trace_script_flag
      ~trace_def:!trace_def_flag
      !targets
  in

  if !mode = `Dump then (
    Godi_make.dump 
      (Godi_file.filename
         ( match !makefile with
	     | None -> user_error "No Makefile available"
	     | Some f -> f
         )
      )
  )
  else (
    let pcache = Godi_make.create_parser_cache() in

    let make_user_file =
      if !exec_mode = `Execute then
        Some (fun file ctx ->
	        let ctx' = { ctx with Godi_make.final_targets = [file] } in
	        Godi_make.execute
		  ~exec_mode:!exec_mode
		  level
		  ctx')
      else
        None
    in
    
    Godi_make.fast_parse_eval
      ~pref_user_error:true
      ~sys_mk:!sys_mk
      ?make_user_file
      pcache
      ( match !makefile with
	  | None -> user_error "No Makefile available"
	  | Some f -> f
      )
      ctx;
    
    if !global_ignore_flag then
      ctx.Godi_make.global_ignore_flag <- true;
    if !global_silent_flag then
      ctx.Godi_make.global_silent_flag <- true;
    
    ( match !mode with
        | `Variable ->
	    List.iter
	      (fun varname ->
	         print_endline (Godi_make.var_definition ctx varname)
	      )
	      !vars
        | `Default | `Target ->
	    Godi_make.execute
	      ~pref_user_error:true
	      ~exec_mode:!exec_mode
	      level
	      ctx
        | _ ->
            ()
    )
  )
;;


let cmd_pax () =
  Pax_cmd.main()
;;


let extra_help_wtool wtool =
  "\n\nThe following URL schemes are supported in this version: \n\n  " ^ 
    String.concat "\n  " wtool#url_help ^ 
    "\n\n" ^ 
    (if wtool # have_auth then
       "Authentication: User names in URLs are respected, but passwords\n\
        are ignored. The special user name 'GODI_USER' is replaced by\n\
        the the user given via the -user option, or, if not set, by\n\
        the contents of the GODI_USER variable set in godi.conf.\n"
     else
       "Only password-less access is supported.\n") ^ 
    "\nOptions:\n"


let char_re = Str.regexp "."


let as_pattern s =
  Str.regexp
    ("^" ^
       (Str.global_substitute
          char_re
          (fun s ->
             match Str.matched_string s with
               | "*" -> "\\([^/]*\\)"
               | other -> Str.quote other
          )
          s
       )
    )


let rec rewrite rewrites url =
  match rewrites with
    | (lhs_pat,rhs) :: rewrites' ->
        if Str.string_match lhs_pat url 0 then
          Str.replace_first
            lhs_pat
            rhs
            url
        else
          rewrite rewrites' url
    | [] ->
        url


let split_rewrite_urls s =
  let words = split s in
  List.map
    (fun word ->
       match decompose_assignment word with
         | Some(lhs,rhs) -> (as_pattern lhs,rhs)
         | None ->
             failwith "Bad -rewrite string"
    )
    words


let cmd_wget () =
  let files = ref [] in
  let timeout = ref 30.0 in
  let user = ref None in
  let wtool = Wtool.wtool_factory() in
  let rewrites = ref [] in
  let outname = ref None in
  Arg.parse 
      [ "-user", Arg.String (fun s -> user := Some s),
	      "<name>            Substitute this user for 'GODI_USER' in authentication";

	"-timeout", Arg.Set_float timeout,
	         "<seconds>      Set the timeout to this value";

        "-rewrite", Arg.String (fun s -> 
                                  rewrites := 
                                    !rewrites @ split_rewrite_urls s),
                 "<lhs>=<rhs>    Replace lhs by rhs";
        "-O", Arg.String (fun s -> outname := Some s),
           "<file>               Output to this file";
      ]
      (fun s -> files := !files @ [s])
      (helptext "[<options>] <url> ..." ^ extra_help_wtool wtool);
(*  let con = stderr_console() in *)
  configure_wtool wtool !timeout !user;
  List.iter
    (fun fn ->
       let fn = rewrite !rewrites fn in
       let tmpdir = Godi_file.allocate_tmpdir() in
       try
	 let dir, base = Godi_file_handler.split_name fn in
	 let wlocation = wtool # open_location dir in
	 let file_r = wlocation # get tmpdir ["/" ^ base] in
	 let file =
	   match file_r with
	     | [ `Ok file ] -> file
	     | [ `Timeout ] -> failwith "Timeout"
	     | [ `Error msg ] -> failwith msg
	     | _ -> assert false in
         let out =
           match !outname with
             | Some n -> n
             | None -> base in
	 Godi_file.copy file (Godi_file.filename out);
	 Godi_file.release_tmpdir tmpdir;
       with
	 | File_handler_error msg ->
	     Godi_file.release_tmpdir tmpdir;
	     raise(User_error("Cannot get " ^ fn ^ ": " ^ msg))
	 | Failure msg ->
	     Godi_file.release_tmpdir tmpdir;
	     raise(User_error("Cannot get " ^ fn ^ ": " ^ msg))
    )
    !files
;;


let make_abs s =
  if s <> "" && s.[0] <> '/' then
    "/" ^ s 
  else
    s


let cmd_wtool () =
  let site = ref "/" in
  let arg1 = ref "" in
  let ops = ref [] in
  let api_args = ref [] in
  let timeout = ref 30.0 in
  let user = ref None in
  let wtool = Wtool.wtool_factory() in
  Arg.parse
    [ "-site", Arg.Set_string site,
            "<url>                       Connect with this site (HTTP, FTP)";

      "-list", Arg.String (fun s -> ops := `List(!site, s) :: !ops),
            "<remotepath>                List files";

      "-get", Arg.Tuple [ Arg.Set_string arg1; 
			  Arg.String (fun s ->
					ops := `Get(!site, !arg1, s) :: !ops )
			],
           "<remotepath> <localpath>     Get file from remote site";

      "-put", Arg.Tuple [ Arg.Set_string arg1; 
			  Arg.String (fun s ->
					ops := `Put(!site, !arg1, s) :: !ops )
			],
           "<localpath> <remotepath>     Put file to remote site";

      "-api", Arg.Rest (fun s ->
                          let (n,v) = 
                            match decompose_assignment s with
                              | Some (n,v) -> (n,v) 
                              | None -> 
                                  failwith ("Bad arg: " ^ s) in
                          if !api_args = [] then
                            ops := `Api !site :: !ops;
                          api_args := (n,v) :: !api_args;
                       ),
           "name=value ...               Call the GODI API with these params";

      "-user", Arg.String (fun s -> user := Some s),
            "<name>                      Substitute this user for 'GODI_USER' in authentication";

      "-timeout", Arg.Set_float timeout,
               "<seconds>                Set the timeout to this value";
    ]
    (fun arg -> raise (Arg.Bad ("Unexpected argument: " ^ arg)))
    (helptext "-site <url> <options>" ^ extra_help_wtool wtool);

  configure_wtool wtool !timeout !user;
  List.iter
    (fun op ->
       match op with
	 | `List(site, remotepath) ->
	     ( try
		 let wlocation = wtool # open_location site in
		 match wlocation # list (make_abs remotepath) with
		   | `Ok names ->
		       let sorted = List.sort String.compare names in
		       List.iter print_endline sorted
		   | `Timeout -> failwith "Timeout"
		   | `Error msg -> failwith msg
	       with
		 Failure msg ->
		   raise(User_error("Cannot get file listing: " ^ msg))
	     )
	 | `Get(site, remotepath, localpath) ->
	     let tmpdir = Godi_file.allocate_tmpdir() in
	     ( try
		 let wlocation = wtool # open_location site in
		 let file_r = wlocation # get tmpdir [make_abs remotepath] in
		 let file =
		   match file_r with
		     | [ `Ok file ] -> file
		     | [ `Timeout ] -> failwith "Timeout"
		     | [ `Error msg ] -> failwith msg
		     | _ -> assert false in
		 Godi_file.copy file (Godi_file.filename localpath);
		 Godi_file.release_tmpdir tmpdir;
 	       with
		 | Failure msg ->
		     Godi_file.release_tmpdir tmpdir;
		     raise(User_error("Cannot download file: " ^ msg))
	     )
	 | `Put(site, localpath, remotepath) ->
	     ( try
		 let wlocation = wtool # open_location site in
		 let r =
		   wlocation # put
		     (Godi_file.filename localpath) 
		     (make_abs remotepath) in
		 match r with
		   | `Ok () -> ()
		   | `Timeout -> failwith "Timeout"
		   | `Error msg -> failwith msg
 	       with
		 | Failure msg ->
		     raise(User_error("Cannot upload file: " ^ msg))
	     )
         | `Api site ->
             ( try
		 let wlocation = wtool # open_location site in
		 let r = wlocation # api (List.rev !api_args) in
                 match r with
                   | `Ok 0 -> ()
                   | `Ok code -> raise(Exit_code code)
                   | `Timeout -> failwith "Timeout"
                   | `Error msg -> failwith msg
 	       with
		 | Failure msg ->
		     raise(User_error("API error: " ^ msg))
             )
    )
    (List.rev !ops)


let rec strip prefixes s =
  let l = String.length s in
  match prefixes with
    | [] ->
        s
    | p :: prefixes' ->
        let n = String.length p in
        if l >= n && String.sub s 0 n = p then
          String.sub s n (l-n)
        else
          strip prefixes' s


let cmd_digest () =
  let alg_s = ref None in
  let files = ref [] in
  let prefixes = ref [] in
  Arg.parse
    [ "-strip", Arg.String (fun s -> prefixes := !prefixes @ [s]),
      "<path>  Strip this part of the path in the output"
    ]
    (fun s ->
       if !alg_s = None then
	 alg_s := Some s
       else
	 files := !files @ [s])
    (helptext "[<options>] (SHA1|MD5) [file ...]");
  
  let alg =
    match !alg_s with
      | Some("SHA1" | "sha1") -> `SHA1
      | Some("MD5" | "md5") -> `MD5
      | Some s ->
	  raise(User_error("Unknown algorithm: " ^ s))
      | None ->
	  raise(User_error("No algorithm on command line")) in

  (* For Win32: otherwise we emit CR chars, and this complicates the
     comparison of checksums
   *)
  set_binary_mode_out stdout true;

  match !files with
    | [] ->
	set_binary_mode_in stdin true;
	let sum = Godi_digest.digest_of_channel alg stdin in
	print_endline (Godi_digest.to_hex sum)
    | _ ->
	List.iter
	  (fun name ->
	     let ch = Godi_file.open_in `Binary (Godi_file.filename name) in
	     let sum = Godi_digest.digest_of_channel alg ch in
             let name' = strip !prefixes name in
	     close_in ch;
	     printf "%s (%s) = %s\n"
	       (match alg with
		  | `SHA1 -> "SHA1"
		  | `MD5 -> "MD5")
	       name'
	       (Godi_digest.to_hex sum)
	  )
	  !files


let cmd_install () =
  let d_flag = ref false in
  let m_flag = ref 0o755 in
  let p_flag = ref false in
  let v_flag = ref false in
  let files = ref [] in
  
  Godi_getopt.parse
    [ "-c", `Unit(fun () -> ()),
      "  (Ignored for compatibilty with other versions of install)";
      
      "-d", `Set d_flag,
      "  Create directories as required";

      "-g", `String (fun _ -> ()),
      "<arg>  (Ignored for compatibilty with other versions of install)";

      "-m", `String (fun m -> m_flag := int_of_string("0o" ^ m)),
      "<mode>  Set an alternate file mode";

      "-o", `String (fun _ -> ()),
      "<arg>  (Ignored for compatibilty with other versions of install)";

      "-p", `Set p_flag,
      "  Preserve the access and modification times";

      "-s", `Unit(fun _ -> ()),
      "  (Ignored for compatibilty with other versions of install)";

      "-v", `Set v_flag,
      "  Set install to being verbose";

      "-h", `Help,
      "  Get help";
    ]
    (fun s -> files := !files @ [Godi_file.filename s])
    (helptext "[<options>] [file/directory...]");
    
  if !d_flag then (
    List.iter
      (fun f ->
	 Godi_file.create_dir ~recurse:true f;
	 Godi_file.set_mode f !m_flag;
	 if !v_flag then
	   eprintf "Created directory %s\n%!" (Godi_file.slashname f)
      )
      !files
  )
  else (
    let install src dest =
      (* copy src to dest *)
      let exe_fixup =
	match Sys.os_type with
	  | "Cygwin" | "Win32" ->
	      not(Filename.check_suffix (Godi_file.basename src) ".exe") &&
		Godi_file.file_exists (Godi_file.addsuffix src ".exe")
	  | _ ->
	      false in
      let (src', dest') =
	if exe_fixup then
	  (Godi_file.addsuffix src ".exe", Godi_file.addsuffix dest ".exe")
	else
	  (src, dest) in
      Godi_file.copy ~remove:true src' dest';
      Godi_file.set_mode dest' !m_flag;
      if !p_flag then
	Godi_file.set_mtime dest' (Godi_file.file_mtime src');
      if !v_flag then
	eprintf "Installed %s\n%!" (Godi_file.slashname dest')
    in

    match !files with
      | [src; dest] when not (Godi_file.file_is_dir dest) ->
	  install src dest
      | [] | [_] ->
	  raise(User_error("At least two arguments required"))
      | _ ->
	  let src_files, dest_dir =
	    match List.rev !files with
	      | d :: s -> (List.rev s), d
	      | _ -> assert false in
	  List.iter
	    (fun src ->
	       let dest = 
		 Godi_file.addname dest_dir (Godi_file.basename src) in
	       install src dest
	    )
	    src_files
  )



let exec_cmd =
  catch_error
    (fun () ->
       let n = Array.length Sys.argv in
       if n < 2 then usage();
       Arg.current := 1;
       match Sys.argv.(1) with
	   "update"  -> cmd_update()
	 | "wish"    -> cmd_wish()
	 | "perform" -> cmd_perform()
	 | "list"    -> cmd_list()
	 | "info"    -> cmd_info()
	 | "dep"     -> cmd_dep()
	 | "delete"  -> cmd_delete()
	 | "add"     -> cmd_add()
	 | "ptool"   -> cmd_ptool()
	 | "profile" -> cmd_profile()
         | "rprofile" -> cmd_rprofile()
         | "initool" -> cmd_initool()
	 | "make"    -> cmd_make 2
	 | "pax"     -> cmd_pax()
	 | "wtool"   -> cmd_wtool()
	 | "wget"    -> cmd_wget()   (* Intentionally undocumented *)
	 | "digest"  -> cmd_digest() (* Intentionally undocumented *)
	 | "install" -> cmd_install()  (* Intentionally undocumented *)
	 | _ ->
	     usage()
    )
;;

